#include "ManipulatorVisualization.h"

//Virtual Robot includes
#include <RobotAPI/libraries/core/remoterobot/RemoteRobot.h>
#include <VirtualRobot/XML/RobotIO.h>
#include <VirtualRobot/Visualization/CoinVisualization/CoinVisualizationFactory.h>

/* Coin includes */
#include <Inventor/actions/SoGetMatrixAction.h>
#include <Inventor/actions/SoSearchAction.h>
#include <Inventor/SbViewportRegion.h>

#include <iostream>

ManipulatorVisualization::ManipulatorVisualization() : isVisualizing(false)
{
    //Im gonna live forever :)
    this->ref();
}

ManipulatorVisualization::~ManipulatorVisualization()
{
    //Life makes no sense anymore
    this->unref();
}

void ManipulatorVisualization::setVisualization(VirtualRobot::RobotPtr robot, VirtualRobot::RobotNodeSetPtr nodeSet)
{
    //Completely forget anything we displayed earlier
    //This also works when this is the first time we display something
    this->removeVisualization();

    //First of all add the end effector
    VirtualRobot::EndEffectorPtr endEffector = robot->getEndEffector(nodeSet->getTCP()->getName());

    if(endEffector)
    {
        VirtualRobot::RobotPtr endEffectorRobot = endEffector->createEefRobot("", "");
        CoinVisualizationPtr endEffectorVisualization = endEffectorRobot->getVisualization<VirtualRobot::CoinVisualization>();
        this->addChild(endEffectorVisualization->getCoinVisualization());
    }
    else
    {
        for(VirtualRobot::RobotNodePtr node : nodeSet->getAllRobotNodes())
        {
            CoinVisualizationPtr endEffectorVisualization = node->getVisualization<VirtualRobot::CoinVisualization>();
            this->addChild(endEffectorVisualization->getCoinVisualization());

            std::cout << "Adding: " << node->getName() << std::endl;
        }
    }
    //And now wrap this bad boy with our manipulator
    this->manip.reset(new SoTransformerManip());

    //We need this null separator to remove the scale knobs
    //This won't lead to memory leak, because this separator
    //will be deleted when the manipulator itself gets removed
    //from the scene graph
    SoSeparator* nullSep = new SoSeparator;

    //Make all scale knobs disappear
    manip->getDragger()->setPart("scale1", nullSep);
    manip->getDragger()->setPart("scale2", nullSep);
    manip->getDragger()->setPart("scale3", nullSep);
    manip->getDragger()->setPart("scale4", nullSep);
    manip->getDragger()->setPart("scale5", nullSep);
    manip->getDragger()->setPart("scale6", nullSep);
    manip->getDragger()->setPart("scale7", nullSep);
    manip->getDragger()->setPart("scale8", nullSep);

    if(endEffector)
    {
        //Move it to the tcp
        Eigen::Matrix4f mat = nodeSet->getTCP()->getGlobalPose();
        mat(0, 3) /= 1000;
        mat(1, 3) /= 1000;
        mat(2, 3) /= 1000;

        //Apply to current and desired visualizer
        manip->setMatrix(VirtualRobot::CoinVisualizationFactory::getSbMatrix(mat));
    }

    this->insertChild(manip.get(), 0);

    this->isVisualizing = true;
}

void ManipulatorVisualization::removeVisualization()
{
    //Remove all children and reset manip pointer
    //This should bring ref counter of Inventor down to zero and free memory
    this->removeAllChildren();
    manip.reset();

    this->isVisualizing = false;
}

void ManipulatorVisualization::addManipFinishCallback(SoDraggerCB *func, void *data)
{
    if(this->getIsVisualizing())
    {
        this->manip->getDragger()->addFinishCallback(func, data);
    }
}

void ManipulatorVisualization::addManipMovedCallback(SoDraggerCB *func, void *data)
{
    if(this->getIsVisualizing())
    {
        this->manip->getDragger()->addMotionCallback(func, data);
    }
}

Eigen::Matrix4f ManipulatorVisualization::getUserDesiredPose()
{
    if(this->getIsVisualizing()) {
        SoGetMatrixAction* action = new SoGetMatrixAction(SbViewportRegion());
        SoSearchAction sa;
        sa.setNode(manip.get());
        sa.setSearchingAll( TRUE );              // Search all nodes
        SoBaseKit::setSearchingChildren( TRUE ); // Even inside nodekits
        sa.apply(this);

        action->apply(sa.getPath());

        SbMatrix matrix = action->getMatrix();

        Eigen::Matrix4f mat = Eigen::Matrix4f::Identity();
        mat (0,0) = matrix[0][0];
        mat (0,1) = matrix[1][0];
        mat (0,2) = matrix[2][0];
        mat (0,3) = matrix[3][0] * 1000;

        mat (1,0) = matrix[0][1];
        mat (1,1) = matrix[1][1];
        mat (1,2) = matrix[2][1];
        mat (1,3) = matrix[3][1] * 1000;

        mat (2,0) = matrix[0][2];
        mat (2,1) = matrix[1][2];
        mat (2,2) = matrix[2][2];
        mat (2,3) = matrix[3][2] * 1000;

        mat (3,0) = matrix[0][3];
        mat (3,1) = matrix[1][3];
        mat (3,2) = matrix[2][3];
        mat (3,3) = matrix[3][3];

        return mat;
    }

    return Eigen::Matrix4f::Identity();
}

std::string ManipulatorVisualization::getUserDesiredPoseString()
{
    Eigen::Matrix4f mat = this->getUserDesiredPose();

    //Convert to string
    std::stringstream buffer;
    buffer << mat;
    return buffer.str();
}



